#include "coordinate_system.h"
#include <QFileDialog>
#include <QSvgGenerator>
#include <QMessageBox>
#define X1 fields[0][0]->value()
#define Y1 fields[0][1]->value()
#define X2 fields[1][0]->value()
#define Y2 fields[1][1]->value()
#define X3 fields[2][0]->value()
#define Y3 fields[2][1]->value()
#define X4 fields[3][0]->value()
#define Y4 fields[3][1]->value()

CoordinateSystem::CoordinateSystem(QWidget *parent) : QWidget(parent) {
    QGridLayout *layout = new QGridLayout();
		QLabel *label = new QLabel("X", this);
		layout->addWidget(label, 0, 0);
		label = new QLabel("Y", this);
		layout->addWidget(label, 0, 1);
		for (char i = 0; i < 10; ++i) {
						(*fields)[i] = new QDoubleSpinBox(this);
						(*fields)[i]->setFixedWidth(buttonsize);
						(*fields)[i]->setRange(-centerX, centerX);
						// fields[i][j]->setDecimals(1);
		}

    for (char i=0;i<4;i++)for (int j = 0; j < 2; ++j) {
        fields[i][j]->setSingleStep(10);
        layout->addWidget(fields[i][j], i+1, j);
        connect(fields[i][j], &QDoubleSpinBox::valueChanged, this, &CoordinateSystem::up);
    }
    QWidget *container = new QWidget(this);
    container->setLayout(layout);
		QPushButton *button = new QPushButton("↻", this);
		QPushButton *revbutton = new QPushButton("↺", this);
		button->setFixedWidth(buttonsize);
		revbutton->setFixedWidth(buttonsize);
     connect(button, &QPushButton::clicked, this, &CoordinateSystem::onButtonClick);
		connect(revbutton, &QPushButton::clicked, this, &CoordinateSystem::onRevButtonClick);
		QLabel *label1 = new QLabel("do around", this);
		layout->addWidget(label1, 0, 2);
		spinBox = new QSpinBox(this);
		spinBox->setFixedWidth(buttonsize);
		layout->addWidget(spinBox, 0, 3);
		spinBox->setRange(1,4);
		QLabel *label2 = new QLabel("A:", this);
		layout->addWidget(label2, 1, 2);
		layout->addWidget(fields[4][1], 1, 3);
		fields[4][1]->setRange(1,10);
		fields[4][1]->setSingleStep(0.1);
		timer = new QTimer(this);
		connect(timer, &QTimer::timeout, this, &CoordinateSystem::rotateStep);	
		layout->addWidget(revbutton, 2, 2);
		layout->addWidget(button, 2, 3);
		
		speedSlider = new QSlider(Qt::Horizontal, this);
    speedSlider->setRange(1, 50);
    speedSlider->setValue(16);
    speedSlider->setInvertedAppearance(true);
    speedSlider->setTickPosition(QSlider::TicksBelow);
    speedSlider->setTickInterval(16);

    QLabel *speedLabel = new QLabel("Speed", this);

    layout->addWidget(speedLabel, 3, 2);
    layout->addWidget(speedSlider, 3, 3);
    connect(speedSlider, &QSlider::valueChanged, this, [this](int value) {
        if(timer->isActive()) {
            timer->start(value);
        }
    });
		warningLabel = new QLabel("Not a parallelogram!", this);
		warningLabel->setStyleSheet("QLabel { background-color : yellow; color : black; font-weight: bold; padding: 5px; }");
		warningLabel->setAlignment(Qt::AlignCenter);
		warningLabel->hide();
		warningLabel->move(0, height() - warningLabel->height());
		QLabel *angleLabel = new QLabel("Angle (deg):", this);
		layout->addWidget(angleLabel, 4, 2);
		angleSpinBox->setRange(-360, 360);
		angleSpinBox->setSingleStep(1);
		angleSpinBox->setValue(90);
		layout->addWidget(angleSpinBox, 4, 3);
		fields[0][0]->setValue(50);
		fields[0][1]->setValue(150);
		fields[1][0]->setValue(150);
		fields[1][1]->setValue(150);
		fields[2][0]->setValue(200);
		fields[2][1]->setValue(250);
		fields[3][0]->setValue(100);
		fields[3][1]->setValue(250);
		for (char i=0;i<4;++i)originalPoints[i]=QPointF(fields[i][0]->value(),fields[i][1]->value());
		QPushButton *saveButton = new QPushButton("Save SVG", this);
		saveButton->setFixedWidth(buttonsize);
		connect(saveButton, &QPushButton::clicked, this, &CoordinateSystem::saveSvg);
		layout->addWidget(saveButton, 7, 0);
		QPushButton *pauseButton = new QPushButton("⏸", this);
		QPushButton *resumeButton = new QPushButton("▶", this);
		
		pauseButton->setFixedWidth(buttonsize);
		resumeButton->setFixedWidth(buttonsize);
		
		layout->addWidget(pauseButton, 7, 2);
		layout->addWidget(resumeButton, 7, 3);
		
		connect(pauseButton, &QPushButton::clicked, this, [this]() {
				if (timer->isActive()) timer->stop();
		});
		
		connect(resumeButton, &QPushButton::clicked, this, [this]() {
				if (!timer->isActive()) timer->start(speedSlider->value());
		});
		QPushButton *resetButton = new QPushButton("↩", this);
		resetButton->setFixedWidth(buttonsize*2+10);
		layout->addWidget(resetButton, 6, 2, 1, 2);
		
		connect(resetButton, &QPushButton::clicked, this, [this]() {
				timer->stop();
				currentStep = 0;
				for (char i = 0; i < 4; ++i) {
						fields[i][0]->setValue(originalPoints[i].x());
						fields[i][1]->setValue(originalPoints[i].y());
				}
		});
		QLabel *stepLabel = new QLabel("Tick size:", this);
		layout->addWidget(stepLabel, 6, 0);
		QSpinBox *stepSpinBox = new QSpinBox(this);
		stepSpinBox->setRange(10, 200);
		stepSpinBox->setSingleStep(10);
		stepSpinBox->setValue(step);
		layout->addWidget(stepSpinBox, 6, 1);
		connect(stepSpinBox, QOverload<int>::of(&QSpinBox::valueChanged), this, [this](int value) {
				step = value;
				update();
		});
		QPushButton *matrixButton = new QPushButton("Matrix", this);
		matrixButton->setFixedWidth(buttonsize);
		layout->addWidget(matrixButton, 7, 1);
		connect(matrixButton, &QPushButton::clicked, this, &CoordinateSystem::showMatrix);
}
bool CoordinateSystem::isParallelogram() {
    QPointF A(X1, Y1);
    QPointF B(X2, Y2);
    QPointF C(X3, Y3);
    QPointF D(X4, Y4);
    QPointF AB = B - A;
    QPointF BC = C - B;
    QPointF CD = D - C;
    QPointF DA = A - D;
    auto isParallel = [](QPointF v1, QPointF v2) {
        float cross = v1.x()*v2.y() - v1.y()*v2.x();
        return std::abs(cross) < 0.01;
    };
    return isParallel(AB, CD) && isParallel(BC, DA);
}

void CoordinateSystem::up() {update();
		if(!currentStep)
		{if (isParallelogram())warningLabel->hide();
		else 	warningLabel->show();
		}}
void CoordinateSystem::paint(QPainter &painter) {
    painter.drawLine(0,centerY,width(),centerY);
    painter.drawLine(centerX,0,centerX,height());
    painter.drawLine(centerX-5,11,centerX,0);
    painter.drawLine(centerX+5,11,centerX,0);
    painter.drawLine(width()-11,centerY-5,width(),centerY);
    painter.drawLine(width()-11,centerY+5,width(),centerY);
    for(short i=centerX;i<width();i+=step)	painter.drawLine(i,centerY-5,i,centerY+5);// 4 цикли - критично.
    for(short i=centerY;i>0;i-=step)				painter.drawLine(centerX-5,i,centerX+5,i);// Можна зробити двома, якщо
    for(short i=centerX;i>0;i-=step)				painter.drawLine(i,centerY-5,i,centerY+5);// рахувати координати не від центру,
    for(short i=centerY;i<height();i+=step)painter.drawLine(centerX-5,i,centerX+5,i);// а від країв.
    painter.setPen(Qt::white);painter.setBrush(Qt::NoBrush);
    painter.drawText(centerX-step+5, centerY - 5, QString::number(-step));
    painter.drawText(centerX+5,centerY - 5 + step, QString::number(-step));
    painter.drawText(centerX+5,centerY - 5 - step, ' ' + QString::number(step));
    painter.drawText(centerX+5,15, "Y");
    painter.drawText(width()-15,centerY-5,"X");
    painter.drawText(centerX+5,centerY-5,"0");
    painter.drawText(centerX+step+5,centerY-5,QString::number(step));
		for (char i=0;i<4;i++){painter.drawEllipse(centerX+fields[i][0]->value()-5, centerY-fields[i][1]->value()-5, 11, 11);
		painter.drawText(centerX+fields[i][0]->value()+5,centerY-fields[i][1]->value()-5,QString::number(i+1));
		painter.drawLine(fields[i][0]->value()+centerX,centerY-fields[i][1]->value(),fields[(i+1)%4][0]->value()+centerX,centerY-fields[(i+1)%4][1]->value());
		}
    painter.end();
}
void CoordinateSystem::paintEvent(QPaintEvent*) {
    QPainter painter(this);
    paint(painter);
		short newMin1 = -centerX;
		if (fields[0][0]->minimum()!=newMin1){
				for (char i=0;i<4;i++){
								fields[i][0]->blockSignals(true);
								fields[i][0]->setRange(newMin1, -newMin1);
								fields[i][0]->blockSignals(false);}
		}
		short newMin = centerY;
		if (fields[0][1]->maximum()!=newMin){
											for (char i=0;i<4;i++){
								fields[i][1]->blockSignals(true);
								fields[i][1]->setRange(-newMin,newMin);
								fields[i][1]->blockSignals(false);}}
}
void CoordinateSystem::onButtonClick() {
    for (char i=0;i<4;++i)
        originalPoints[i]=QPointF(fields[i][0]->value(),fields[i][1]->value());
    reverseRotation=false;
    targetScale=fields[4][1]->value();targetAngle = angleSpinBox->value()*M_PI/180;
    timer->start(speedSlider->value());
}

void CoordinateSystem::onRevButtonClick() {
    for (char i=0;i<4;++i)
        originalPoints[i]=QPointF(fields[i][0]->value(),fields[i][1]->value());
    reverseRotation=true;
    targetScale=1/fields[4][1]->value();targetAngle=angleSpinBox->value()*M_PI/180;
    timer->start(speedSlider->value());
}

void CoordinateSystem::rotateStep() {
    #define spinIndex spinBox->value()-1
	// for (char i = 0; i < 8; ++i) (*fields)[i]->setDecimals(2);
#define totalSteps 62
    float progress=float(currentStep)/totalSteps;
    float effectiveAngle=(reverseRotation?-targetAngle:targetAngle)*progress;
    float currentScale=1.0+(targetScale-1.0)*progress;

    for (char i=0;i<4;++i) {
        float dx=originalPoints[i].x()-originalPoints[spinIndex].x();
        float dy=originalPoints[i].y()-originalPoints[spinIndex].y();
        float newX=(cos(effectiveAngle)*dx+sin(effectiveAngle)*dy)/currentScale+originalPoints[spinIndex].x();
        float newY=(-sin(effectiveAngle)*dx+cos(effectiveAngle)*dy)/currentScale+originalPoints[spinIndex].y();
        fields[i][0]->setValue(newX);
        fields[i][1]->setValue(newY);
    }
    currentStep++;
    if (currentStep>totalSteps) {
        timer->stop();
				currentStep=0;
				// for (char i = 0; i < 8; ++i) (*fields)[i]->setDecimals(1);
        return;
    }
}
void CoordinateSystem::saveSvg() {
    QString path=QFileDialog::getSaveFileName(this,tr("Save as SVG"),"",tr("SVG file (*.svg)"));
    if (path.isEmpty())return;

    QSvgGenerator generator;
    generator.setFileName(path);
    generator.setSize(QSize(width(),height()));
    generator.setViewBox(QRect(0,0,width(),height()));

    QPainter painter;
    painter.begin(&generator);
    paint(painter);
    painter.end();
}

void CoordinateSystem::showMatrix() {
    QMessageBox::information(this, "Final matrix", QString(
			"<pre>"
			"[%1, %2, 1]\n"
			"[%3, %4, 1]\n"
			"[%5, %6, 1]\n"
			"[%7, %8, 1]"
			"</pre>")
			.arg(QString::asprintf("%6.2f", X1))
			.arg(QString::asprintf("%6.2f", Y1))
			.arg(QString::asprintf("%6.2f", X2))
			.arg(QString::asprintf("%6.2f", Y2))
			.arg(QString::asprintf("%6.2f", X3))
			.arg(QString::asprintf("%6.2f", Y3))
			.arg(QString::asprintf("%6.2f", X4))
			.arg(QString::asprintf("%6.2f", Y4))
			);
}
